// Copyright 2005,2006,2007 Markus Schordan, Gergo Barany
// $Id: cfg_support.C,v 1.10 2008-03-05 17:08:26 gergo Exp $

#include "CFGTraversal.h"
#include "cfg_support.h"
#include "IrCreation.h"

CallBlock::CallBlock(KFG_NODE_ID id_, KFG_NODE_TYPE type_, int procnum_,
		     std::vector<SgVariableSymbol *> *paramlist_, const char *name_)
  : BasicBlock(id_, type_, procnum_), paramlist(paramlist_), name(name_)
{
  switch (node_type) {
  case X_FunctionEntry:
    statements.push_back(stmt = Ir::createFunctionEntry(node_type, name, this));
    break;
  case X_FunctionExit:
    statements.push_back(stmt = Ir::createFunctionExit(node_type, name, this));
    break;
  default:
    statements.push_back(stmt = Ir::createCallStmt(node_type, name, this));
    }
}

std::string DeclareStmt::unparseToString() const 
{
  std::string label = "DeclareStmt(";
  label += var->get_name().str();
  label += ", ";
  label += Ir::fragmentToString(type);
  label += ")";
  return label;
}

std::string UndeclareStmt::unparseToString() const 
{
  std::string label = "UndeclareStmt([";
  if (vars != NULL) {
    std::vector<SgVariableSymbol *>::const_iterator i;
    i = vars->begin();
    if (i != vars->end()) {
      label += (*i++)->get_name().str();
    }
    while (i != vars->end()) {
      label += ", ";
      label += (*i++)->get_name().str();
    }
  }
  label += "])";
  return label;
}

// GB (2007-10-23): Added this function to unparse the new members of the
// ExternalCall node (the function expression and the list of parameter
// variables).
std::string ExternalCall::unparseToString() const
{
    std::stringstream label;
    label << "ExternalCall(" << expr_to_string(function) << ", [";
    assert(params != NULL);
    std::vector<SgVariableSymbol *>::const_iterator i = params->begin();
    if (i != params->end())
    {
        label << (*i)->get_name().str();
        for (++i; i != params->end(); ++i)
            label << ", " << (*i)->get_name().str();
    }
    label << "])";
    return label.str();
}

std::string CallBlock::print_paramlist() const 
{
  if (paramlist == NULL) {
    // return std::string("null pointer in ") + (void *) this;
    std::stringstream s;
    s << "CallBlock " << (void *) this << " has null params";
    return s.str();
  }
  std::vector<SgVariableSymbol *>::const_iterator i = paramlist->begin();
  if (i == paramlist->end()) {
    return std::string("");
  } else {
    std::stringstream s;
    s << (*i)->get_name().str();
    for (++i; i != paramlist->end(); ++i)
      s << ", " << (*i)->get_name().str();
    return s.str();
  }
}

CallStmt::CallStmt(KFG_NODE_TYPE type_, const char *name_, CallBlock *parent_)
  : type(type_), name(name_), parent(parent_) 
{
  update_infolabel();
}

void 
CallStmt::update_infolabel() 
{
  std::string s;
  switch (type) {
  case X_FunctionCall:
    s = std::string("Call(") + name +
      + ", [" + parent->print_paramlist() + "])";
    break;
  case X_FunctionReturn:
    s = std::string("Return(") + name +
      + ", [" + parent->print_paramlist() + "])";
    break;
  case X_FunctionEntry:
    s = std::string("Entry(") + name + ")";
    break;
    case X_FunctionExit:
      s = std::string("Exit(") + name
	+ ", [" + parent->print_paramlist() + "])";
        break;
  default:
    s = "WEIRD NODE";
  }
  infolabel = s;
}

const char*
expr_to_string(const SgExpression *expr) 
{
  if (expr == NULL)
    return "null";

  // MS: deactivate code because new function fragmentToString allow
  // to unparse without duplicating code
  //SgTreeCopy treecopy;
  //SgExpression *new_expr = isSgExpression(expr->copy(treecopy));
  //new_expr->set_parent(NULL);
  // const char* retval = strdup(Ir::fragmentToString(new_expr).c_str());

  const char* retval = strdup(Ir::fragmentToString(expr).c_str());
  return retval;
}

bool ExprPtrComparator::operator()(const SgExpression *a,
				   const SgExpression *b) const 
{
  const char *sa = expr_to_string(a), *sb = expr_to_string(b);
  return strcmp(sa, sb) < 0;
}

bool TypePtrComparator::operator()(SgType *a, SgType *b) const 
{
  while (isSgTypedefType(a)) {
    a = isSgTypedefType(a)->get_base_type();
  }
  while (isSgTypedefType(b)) {
    b = isSgTypedefType(b)->get_base_type();
  }
  const char *sa = strdup(Ir::fragmentToString(a).c_str());
  const char *sb = strdup(Ir::fragmentToString(b).c_str());
  
  return strcmp(sa, sb) < 0;
}

SgFunctionRefExp*
find_called_func(SgExpression *call_expr)
{
  if (isSgFunctionRefExp(call_expr))
    return isSgFunctionRefExp(call_expr);
  else if (isSgDotExp(call_expr))
    return find_called_func(isSgDotExp(call_expr)->get_rhs_operand());
  else if (isSgArrowExp(call_expr))
    return find_called_func(isSgArrowExp(call_expr)->get_rhs_operand());
  else
    return NULL;
}

SgMemberFunctionRefExp*
find_called_memberfunc(SgExpression *call_expr)
{
  if (isSgMemberFunctionRefExp(call_expr))
    return isSgMemberFunctionRefExp(call_expr);
  else if (isSgDotExp(call_expr))
    return find_called_memberfunc(isSgDotExp(call_expr)->get_rhs_operand());
  else if (isSgArrowExp(call_expr))
    return find_called_memberfunc(isSgArrowExp(call_expr)->get_rhs_operand());
  else
    return NULL;
}

SgExpression*
calling_object_address(SgExpression *call_expr)
{
  SgDotExp *dot_exp = isSgDotExp(call_expr);
  SgArrowExp *arrow_exp = isSgArrowExp(call_expr);
  SgMemberFunctionRefExp *mfun_ref = isSgMemberFunctionRefExp(call_expr);
  
  if (dot_exp)
  {
    if (isSgMemberFunctionRefExp(dot_exp->get_rhs_operand()))
      return Ir::createAddressOfOp(dot_exp->get_lhs_operand(), dot_exp->get_type());
    else
      return calling_object_address(dot_exp->get_rhs_operand());
  }
  else if (arrow_exp)
  {
    if (isSgMemberFunctionRefExp(arrow_exp->get_rhs_operand()))
      return arrow_exp->get_lhs_operand();
    else
      return calling_object_address(arrow_exp->get_rhs_operand());
  }
  else if (mfun_ref)
  {
    return Ir::createNullPointerExp(
            new SgPointerType(mfun_ref->get_symbol()
                ->get_declaration()->get_associatedClassDeclaration()
                ->get_type()));
  }
  else
    return NULL;
}

std::string *
find_func_name(SgFunctionCallExp *call)
{
  SgFunctionRefExp *fr = find_called_func(call->get_function());
  SgMemberFunctionRefExp *mfr = find_called_memberfunc(call->get_function());
  
  if (fr)
    return new std::string(fr->get_symbol_i()->get_name().str());
  else if (mfr)
    return new std::string(mfr->get_symbol_i()->get_name().str());
  else
    return NULL;
}

std::string
CallStmt::unparseToString() const
{
  return infolabel;
}

const char*
CallStmt::get_funcname() const
{
  return name;
}

std::string IcfgStmt::unparseToString() const
{
  std::cout << "IcfgStmt::UnparseToString: use inherited class." << std::endl;
  assert(false);
}

ArgumentAssignment::ArgumentAssignment(SgVariableSymbol *l, SgExpression *r)
{
  init(Ir::createVarRefExp(l), r);
}

ArgumentAssignment::ArgumentAssignment(SgExpression *l, SgExpression *r)
{
  init(l, r);
}

void ArgumentAssignment::init(SgExpression *l, SgExpression *r)
{
  lhs = l;
  if (isSgFunctionCallExp(r)) {
    RetvalAttribute *varnameattr
      = (RetvalAttribute *) r->getAttribute("return variable");
    rhs = Ir::createVarRefExp(varnameattr->get_str(),r->get_type());
  } else if (isSgConstructorInitializer(r)
	     && r->attributeExists("anonymous variable")) {
    RetvalAttribute *varnameattr
      = (RetvalAttribute *) r->getAttribute("anonymous variable");
    rhs = Ir::createVarRefExp(varnameattr->get_str(),r->get_type());
  } else {
    rhs = r;
  }
  if (rhs != NULL) {
 // rhs->set_parent(NULL);
  }
}

SgExpression *ArgumentAssignment::get_lhs() const
{
  return lhs;
}

SgExpression *ArgumentAssignment::get_rhs() const
{
  return rhs;
}

void ArgumentAssignment::set_rhs(SgExpression *r)
{
  rhs = r;
}

std::string ArgumentAssignment::unparseToString() const
{
  std::string buf = "ArgumentAssignment(";
  buf += expr_to_string(lhs);
  buf += ", ";
  buf += expr_to_string(rhs);
  buf += ")";
  return buf;
}

std::string ReturnAssignment::unparseToString() const
{   
  std::string buf = "ReturnAssignment("
    + Ir::getStrippedName(lhs->get_declaration()) 
    + ", "
    + Ir::getStrippedName(rhs->get_declaration())
    + ")";
  return buf;
} 

std::string ParamAssignment::unparseToString() const
{   
  std::string buf = "ParamAssignment("
    + Ir::getStrippedName(lhs->get_declaration()) 
    + ", "
    + Ir::getStrippedName(rhs->get_declaration());
    + ")";
  return buf;
} 

std::string LogicalIf::unparseToString() const
{
  return std::string("LogicalIf(") + Ir::fragmentToString(expr) + ')';
}

std::string IfJoin::unparseToString() const
{
  return std::string("IfJoin");
}

std::string WhileJoin::unparseToString() const
{
  return std::string("WhileJoin");
}

std::string ConstructorCall::unparseToString() const
{
  return std::string("ConstructorCall(") + get_name() + ")";
}

std::string DestructorCall::unparseToString() const
{
  return std::string("DestructorCall(") + get_name() + ")";
}

SgVariableSymbol * MyAssignment::get_lhs() const {
  return lhs;
}

SgVariableSymbol * MyAssignment::get_rhs() const {
  return rhs;
}

MyAssignment::MyAssignment(SgVariableSymbol *l, SgVariableSymbol *r) : lhs(l), rhs(r) {
}

BasicBlock *call_destructor(SgInitializedName *in, CFG *cfg,
			    int procnum, BasicBlock *after, int *node_id)
{
  SgClassType *ct = isSgClassType(in->get_type());
  std::string class_name(ct->get_name().str());
  std::string destructor_name = class_name + "::~" + class_name;
  std::string this_var_name = std::string("$~") + class_name + "$this";
  std::deque<Procedure *>::const_iterator i;
  for (i = cfg->procedures->begin(); i != cfg->procedures->end(); ++i) {
    /* we want member functions */
    if ((*i)->memberf_name == NULL)
      break;
    if (destructor_name == (*i)->memberf_name) {
      SgVariableSymbol* this_var_sym = Ir::createVariableSymbol(this_var_name, in->get_type());

      CallBlock *entry = (*i)->entry;
      CallBlock *call_block = new CallBlock((*node_id)++, CALL,
					    procnum, NULL, strdup(destructor_name.c_str()));
      CallBlock *return_block = new CallBlock((*node_id)++, RETURN,
					      procnum, NULL, strdup(destructor_name.c_str()));
      cfg->nodes.push_back(call_block);
      cfg->calls.push_back(call_block);
      cfg->nodes.push_back(return_block);
      cfg->returns.push_back(return_block);
      call_block->partner = return_block;
      return_block->partner = call_block;
      BasicBlock *this_block
	= new BasicBlock((*node_id)++, INNER, procnum);
      cfg->nodes.push_back(this_block);

      SgAddressOfOp* addressOfOp
	= Ir::createAddressOfOp(Ir::createVarRefExp(Ir::createVariableSymbol(in)),
				Ir::createPointerType(in->get_type()));
      ArgumentAssignment* argumentAssignment
	= new ArgumentAssignment(Ir::createVarRefExp(this_var_sym),addressOfOp);
      this_block->statements.push_back(argumentAssignment);

      /* set links */
      add_link(this_block, call_block, NORMAL_EDGE);
      add_link(call_block, entry, CALL_EDGE);
      add_link(call_block, return_block, LOCAL);
      add_link(entry->partner, return_block, RETURN_EDGE);
      add_link(return_block, after, NORMAL_EDGE);
      after = this_block;
      /* This was the only implementation of this
       * destructor; we don't need to traverse the rest of
       * the procedures. */
      break;
    }
  }
  /* If no destructor was found, it is external -- or does not
   * exist, this should be improved in the future. */
  if (i == cfg->procedures->end()) {
    BasicBlock *b = new BasicBlock((*node_id)++, INNER, procnum);
    cfg->nodes.push_back(b);
    b->statements.push_back(Ir::createDestructorCall(strdup(class_name.c_str()), ct));
    add_link(b, after, NORMAL_EDGE);
    after = b;
  }
  return after;
}

bool subtype_of(SgClassDefinition *a, SgClassDefinition *b)
{
  if (a == NULL || b == NULL)
      return false;
  const SgBaseClassPtrList &base_classes = a->get_inheritances();
  SgBaseClassPtrList::const_iterator i;
  for (i = base_classes.begin(); i != base_classes.end(); ++i) {
    SgClassDefinition *base = (*i)->get_base_class()->get_definition();
    if (base == b || subtype_of(base, b)) {
      return true;
    }
  }
  return false;
}

void dumpTreeFragment(SgNode *node, std::ostream &stream)
{
    TreeFragmentDumper tfd(stream);
    tfd.run(node);
}

std::string dumpTreeFragmentToString(SgNode *node)
{
    std::stringstream ss;
    dumpTreeFragment(node, ss);
    return ss.str();
}
